#include "Color.h"
#include "Image.h"
#include "Ray.h"
#include "Sphere.h"
#include "Box.h"
#include "Triangle.h"
//#include "BoundingVolumeHierarchy_opt.h" // full optimization isn't as fast as just the closer hit check
#include "BoundingVolumeHierarchy.h"
#include "PinholeCamera.h"
#include "PointLight.h"
#include "LambertianMaterial.h"
#include "trax.hpp"

int main()
{
  trax_setup();

  BoundingVolumeHierarchy bvh(loadi(0, 8)); // tell the BVH where the scene starts
  PointLight light = loadLightFromMemory(loadi(0, 12)); // load the light

  int xres = loadi( 0, 1 );
  int yres = loadi( 0, 4 ); 
  float inv_width = loadf(0, 2);
  float inv_height = loadf(0, 5);
  int start_fb = loadi( 0, 7 );
  int start_matls = loadi(0, 9);
  //int num_samples = loadi(0, 17);
  //int max_depth = loadi(0, 16);
  int num_samples = 1; // hard-coded for run-rt
  int max_depth = 1;
 
  // set up frame buffer
  Image image(xres, yres, start_fb);
  
  // load camera
  PinholeCamera camera(loadi(0, 10));

  // loop over pixels
  for(int pix = atomicinc(0); pix < xres*yres; pix = atomicinc(0))
    {
      //trax_printi(pix);
      int i = pix / xres;
      int j = pix % xres;

      Color result(0.f, 0.f, 0.f);
      
      // find center of pixel
      float x = 2.0f * (j - xres/2.f + 0.5f)/xres;
      float y = 2.0f * (i - yres/2.f + 0.5f)/yres;

      for(int i=0; i < num_samples; i++)
	{
	  Color attenuation(1.f, 1.f, 1.f);
	  int depth = 0;
	  Ray ray;

	  // cast a ray for this pixel
	  if(num_samples == 1)
	    camera.makeRay(ray, x, y);
	  else // randomize ray within pixel
	    {
	      float x_off = (trax_rand() - 0.5f) * 2.f;
	      float y_off = (trax_rand() - 0.5f) * 2.f;
	      
	      x_off *= inv_width;
	      y_off *= inv_height;
	      camera.makeRay(ray, x + x_off, y + y_off);
	    }
	  // let ray bounce max_depth-1 times (camera ray counts)
	  while(depth++ < max_depth)
	    {	      
	      HitRecord hit(100000.f);
	      Color hit_color;
	      Vector normal;
	      Vector hit_point;
	      
	      // intersect the ray with the scene
	      bvh.intersect(hit, ray);
	      
	      // Add light for this ray
	      result += shade(hit, ray, bvh, light, start_matls, hit_point, normal, hit_color) * attenuation;
	      
	      // can't continue the ray path if we hit the background
	      if(!hit.didHit())
		break;

	      // Set up ray for next bounce
	      ray.org = hit_point;
	      ray.dir = randomReflection(normal);
	      attenuation *= hit_color; // path loses energy on each bounce
	    }

	}
      result /= num_samples;
      // apply a tone-map here if desired
      image.set(i, j, result);
  }
  trax_cleanup();
}
